/*
    MLIST.CPP
    ---------
    Thread-safe generic linked-list package
*/

#include "wpch.hpp"
#include "mlist.hpp"


/*
 * MListNew -- Start a new list.  Warning: clobbers over any existing
 *         data and resets the semaphores...
 */

WBool MListNew( mlist_header_p top )
{
    WBool		ok = FALSE;
    WCriticalSection	critSect;
    WSemaphoreLock	lock( &critSect );

    if( !top ) goto leave;
    top->mutex = new WMutexSemaphore( NULL );
    top->data = new WEventSemaphore( NULL );

    top->first = NULL;
    top->last  = NULL;

    ok = TRUE;

leave:
    return( ok );
}

/*
 * MListLock -- Grab the mutex semaphore to lock the list.
 */

WBool MListLock( mlist_header_p top )
{
    WBool ok = FALSE;

    if( top ) {
	ok = top->mutex->Acquire();
    }
    return( ok );
}


/*
 * MListUnlock -- Release a locked list.
 */

WBool MListUnlock( mlist_header_p top )
{
    WBool ok = FALSE;

    if( top ) {
	ok = top->mutex->Release();
    }
    return( ok );
}


/*
 * MListIsEmpty -- Returns TRUE if the list is empty...
 */

WBool MListIsEmpty( mlist_header_p top )
{
    WBool empty = TRUE;

    if( top ) {
	if( top->mutex->Acquire() ) {
	    empty = ( top->first == NULL );
	    top->mutex->Release();
	}
    }
    return( empty );
}


/*
 * MListPrefix -- Add an entry to the start of a list.
 */

WBool MListPrefix( mlist_header_p top, void *entry )
{
    WBool ok = FALSE;

    if( !top || !entry ) goto leave;
    if( !top->mutex->Acquire() ) goto leave;

    ((mlist_link_p) entry)->prev = NULL;
    ((mlist_link_p) entry)->next = top->first;

    if( top->first )
        top->first->prev = (mlist_link_p) entry;

    top->first = (mlist_link_p) entry;

    if( !top->last )
        top->last = (mlist_link_p) entry;

    #ifdef DEBUG_MLIST
        if( top->first == NULL || top->last == NULL ){
            Dbg(( "Error in MListPrefix\n" ));
        }
    #endif

    top->data->Set();
    top->mutex->Release();
    ok = TRUE;

leave:
    return( ok );
}


/*
 * MListSuffix -- Add an entry to the end of a list.
 */

WBool MListSuffix( mlist_header_p top, void *entry )
{
    WBool ok = FALSE;

    if( !top || !entry ) goto leave;
    if( !top->mutex->Acquire() ) goto leave;

    ((mlist_link_p) entry)->prev = top->last;
    ((mlist_link_p) entry)->next = NULL;

    if( top->last )
        top->last->next = (mlist_link_p) entry;

    top->last = (mlist_link_p) entry;

    if( !top->first )
        top->first = (mlist_link_p) entry;

    #ifdef DEBUG_MLIST
        if( top->first == NULL || top->last == NULL ){
            Dbg(( "Error in MListSuffix\n" ));
        }
    #endif

    top->data->Set();
    top->mutex->Release();
    ok = TRUE;

leave:
    return( ok );
}


/*
 * MListDelete -- Remove an entry from the list.  Returns the entry.
 */

void *MListDelete( mlist_header_p top, void *entry )
  {
    void *deleted = NULL;

    #ifdef DEBUG_MLIST
    bool          test;
    mlist_link_p  curr;
    #endif

    if( !top || !entry ) goto leave;
    if( !top->mutex->Acquire() ) goto leave;

    #ifdef DEBUG_MLIST
        for( curr = top->first; curr != NULL; curr = curr->next ){
            if( curr == entry )
                break;
        }

        if( !curr ){
            Dbg(( "Error in MListDelete (not in list!)\n" ));
        }
    #endif

    if( ((mlist_link_p) entry)->next )
        ((mlist_link_p) entry)->next->prev = ((mlist_link_p) entry)->prev;

    if( ((mlist_link_p) entry)->prev )
        ((mlist_link_p) entry)->prev->next = ((mlist_link_p) entry)->next;

    #ifdef DEBUG_MLIST
        test = ( top->first == top->last );
    #endif

    if( entry == (void *) top->first )
        top->first = ((mlist_link_p) entry)->next;

    if( entry == (void *) top->last )
        top->last = ((mlist_link_p) entry)->prev;

    #ifdef DEBUG_MLIST
        if( test ){
            if( top->first != NULL || top->last != NULL ){
                Dbg(( "Error in MListDelete (bad list header)\n" ));
            }
        }
    #endif

    ((mlist_link_p) entry)->next = NULL;
    ((mlist_link_p) entry)->prev = NULL;

    /* If list is now empty, clear the data semaphore */
    if( !top->first ){
        top->data->Reset();
    }

    top->mutex->Release();

    deleted = entry;

leave:
    return( deleted );
}


/*
 * MListInsert -- Insert an entry into a list.
 */

WBool MListInsert( mlist_header_p top, void *next, void *entry )
{
    WBool ok = FALSE;

    if( !top || !entry ) goto leave;
    if( !top->mutex->Acquire() ) goto leave;

    if( top->first == NULL || top->first == next ){
        ok = MListPrefix( top, entry );
    } else if( next == NULL ){
        ok = MListSuffix( top, entry );
    } else {
        ((mlist_link_p) entry)->prev       = ((mlist_link_p) next)->prev;
        ((mlist_link_p) entry)->next       = (mlist_link_p) next;

        WASSERT( ((mlist_link_p) next)->prev != NULL );

        ((mlist_link_p) next)->prev->next  = (mlist_link_p) entry;
        ((mlist_link_p) next)->prev    = (mlist_link_p) entry;

        ok = TRUE;
    }

    #ifdef DEBUG_MLIST
        if( top->first == NULL || top->last == NULL ){
            Dbg(( "Error in MListInsert\n" ));
        }
    #endif

    top->data->Set();
    top->mutex->Release();

leave:
    return( ok );
}


/*
 * MListDestroy -- Destroy a list's semaphores and other allocated memory
 */

WBool MListDestroy( mlist_header_p top )
{
    WBool     	ok = FALSE;
    WCriticalSection	critSect;
    WSemaphoreLock	lock( &critSect );

    if( !top ) goto leave;

    if( top->data ) {
	top->data->Set();
	delete top->data;
	top->data = NULL;
    }
    if( top->mutex ) {
	delete top->mutex;
	top->mutex = NULL;
    }
    top->first = NULL;
    top->last  = NULL;

    ok = TRUE;

leave:
    return( ok );
}



/*
 * MListKill -- Remove an entire list.
 */

WBool MListKill( mlist_header_p top, void (*fn_free)( void * ) )
{
    WBool     	ok = FALSE;
    mlist_link_p curr = NULL, next = NULL;

    if( !top || !fn_free ) goto leave;
    if( !top->mutex->Acquire() ) goto leave;

    for( curr = top->first; curr != NULL; curr = next ) {
        next = curr->next;
        (*fn_free)( (void *) curr );
    }

    top->first = NULL;
    top->last  = NULL;
    top->mutex->Release();

    ok = TRUE;

leave:
    return( ok );
}

