/*
    mlist.h
    -------
    External definitions for mlist.cpp (Multithread generic linked
    list)
*/

#ifndef MLIST_HPP_INCLUDED
#define MLIST_HPP_INCLUDED


    /* Generic list link field */

    typedef struct mlist_link   *mlist_link_p;
    typedef struct mlist_header *mlist_header_p;

    typedef struct mlist_link {
        mlist_link_p prev, next;
    } mlist_link;

    typedef struct mlist_header {
	WMutexSemaphore *mutex; /* Use to control access to list */
        WEventSemaphore *data;  /* Posted when data is added to the list */
        mlist_link_p first;
        mlist_link_p last;
    } mlist_header;

    #define MLNext( x )       (void *)(((mlist_link_p)(x))->next)
    #define MLPrev( x )       (void *)(((mlist_link_p)(x))->prev)
    #define MLFirst( x )      (void *)(((mlist_header_p)(x))->first)
    #define MLLast( x )       (void *)(((mlist_header_p)(x))->last)

    #define MListRemoveFirst(x)   (void *)MListDelete(x,MLFirst(x))

    extern WBool    MListIsEmpty( mlist_header_p top );
    extern WBool    MListNew( mlist_header_p top );
    extern WBool    MListLock( mlist_header_p top );
    extern WBool    MListUnlock( mlist_header_p top );
    extern WBool    MListPrefix( mlist_header_p top, void * entry );
    extern WBool    MListSuffix( mlist_header_p top, void * entry );
    extern void *   MListDelete( mlist_header_p top, void * entry );
    extern WBool    MListInsert( mlist_header_p top, void * prev,
                                 void * entry );
    extern WBool    MListKill( mlist_header_p top, void (*free)( void * ) );
    extern WBool    MListDestroy( mlist_header_p top );


template<class T>
class MList {

    public:
	MList() { MListNew( &_header ); }
	~MList() { MListDestroy( &_header ); }

	WBool Lock() { return( MListLock( &_header ) ); }
	WBool Unlock() { return( MListUnlock( &_header ) ); };
	WBool WaitForItem( unsigned long timeout )
			{ return( _header.data->Wait( timeout ) ); }
	WBool WaitForItem() { return( _header.data->Wait() ); }
	void Prefix( T *item ) { MListPrefix( &_header, item ); }
	void Suffix( T *item ) { MListSuffix( &_header, item ); }
	void Delete( T *item ) { MListDelete( &_header, item ); }
	void Kill() { MListKill( &_header, &MList::killItem ); }
	
	T *First() { return( (T *)MLFirst( &_header ) ); }
	T *Next( T *item ) { return( (T *)MLNext( item ) ); }
	T *Prev( T *item ) { return( (T *)MLPrev( item ) ); }
	T *Last() { return( (T *)MLLast( &_header ) ); }

	mlist_header    _header;
	
    private:
	static void killItem( void *p ) {
	    T *p2 = (T *) p;
	    delete p2;
	};
};


class WCMDEF MListLink : public mlist_link {
    public:
	mlist_link *GetLink( void ) { return( this ); };
};


template<class T>
class MListClass : public MList<mlist_link> {

    public:
	void Prefix( T *item ) { MList::Prefix( item->GetLink() ); }
	void Suffix( T *item ) { MList::Suffix( item->GetLink() ); }
	void Delete( T *item ) { MList::Delete( item->GetLink() ); }
	void Kill() { MListKill( &_header, &MListClass::killItem ); }
	
	T *First() { return( (T *)MList::First() ); };
	T *Next( T *item ) { return( (T *)MList::Next( item->GetLink() ) ); };
	T *Prev( T *item ) { return( (T *)MList::Prev( item->GetLink() ) ); };
	T *Last() { return( (T *)MList::Last() ); };
	
    private:
	static void killItem( void *p ) {
		MListLink 		*link_item;
		T			*item;
		
		link_item = (MListLink *)p;
		item = (T *)link_item;
		delete item;
	    };
};




#endif 
