#ifndef _DPCLIENT_H_INCLUDED
#define _DPCLIENT_H_INCLUDED

#ifndef _WINDOWS_
#include "windows.h"
#endif

#ifdef __cplusplus
extern "C" {
#endif

#pragma pack(push,4);

#define DPCLIENT_TYPE_INTERNAL    0x0000
#define DPCLIENT_TYPE_INIT        0x0001
#define DPCLIENT_TYPE_FINI        0x0002
#define DPCLIENT_TYPE_STRING      0x0004
#define DPCLIENT_TYPE_THREADNAME  0x0008
#define DPCLIENT_TYPE_CONCAT      0x8000
#define DPCLIENT_TYPE_MEMORY      0x4000
#define DPCLIENT_TYPE_USER        0x2000
#define DPCLIENT_TYPE_THREAD      0x1000
#define DPCLIENT_TYPE_SYSTEM      0x0800
#define DPCLIENT_TYPE_LIBRARY     0x0400
#define DPCLIENT_TYPE_DEBUG       0x0200

//
// Maximum message is has to < 400 bytes.
//

#define _MAX_DPMESSAGE_SIZE 256
#define _DP_TIMEOUT 100

typedef struct _dpclient_message_header {
    unsigned short zeroArea;    // set it to all zeroes
    unsigned short size;        // size of this struct
    unsigned short messageSize; // number of bytes in message part
    unsigned short type;        // message type
    unsigned long  pid;         // from process #
    unsigned long  tid;         // from thread #
    unsigned long  messageID;   // unique ID from that machine
    char           machineID[ 50 ]; // machine name
} dpclient_message_header;

typedef struct _dpclient_message {
    dpclient_message_header header;
    unsigned char           data[ _MAX_DPMESSAGE_SIZE ];
} dpclient_message;

extern HANDLE DPOpen( const char *serverName );
extern BOOL   DPThreadName( HANDLE handle, const char *name );
extern BOOL   DPConnected( void );
extern BOOL   DPSendString( HANDLE handle, const char *string );
extern BOOL   DPSendTypedString( HANDLE handle, const char *string, USHORT type );
extern BOOL   DPClose( HANDLE handle );
extern BOOL   DPWrite( HANDLE handle, dpclient_message *msg );
extern void   DPFillHeader( dpclient_message_header *header,
                            unsigned short type );

#ifdef DEFINE_DPCLIENT

static HANDLE           _DPSendTo = 0;
static CRITICAL_SECTION _DPCrit;
static DWORD            _DPNext = 0;

void DPInit()
{
    if( _DPNext == 0 ){
	InitializeCriticalSection( &_DPCrit );
	_DPNext = GetTickCount();
	if( _DPNext == 0 ){
	    _DPNext = 1000;
	}
    }
}

void DPFillHeader( dpclient_message_header *header,
                   unsigned short type )
{
    if( _DPNext == 0 ){
	DPInit();
    }

    EnterCriticalSection( &_DPCrit );

    header->zeroArea     = 0;
    header->size         = sizeof( *header );
    header->pid          = GetCurrentProcessId();
    header->tid          = GetCurrentThreadId();
    header->machineID[0] = 0;

    DWORD cbWritten = 50;
    GetComputerName( header->machineID, &cbWritten );

    header->type         = type;
    header->messageSize  = 0;
    header->messageID    = _DPNext++;

    LeaveCriticalSection( &_DPCrit );
}

HANDLE DPOpen( const char *serverName )
{
    HANDLE slot;

    if( _DPNext == 0 ){
	DPInit();
    }

    if( !serverName ){
	serverName = "\\\\.\\mailslot\\drainpip";
    }

    EnterCriticalSection( &_DPCrit );

    slot = CreateFile( serverName, GENERIC_WRITE,
                       FILE_SHARE_READ, NULL, OPEN_EXISTING,
	               FILE_ATTRIBUTE_NORMAL, NULL );
    if( slot == 0 || slot == INVALID_HANDLE_VALUE ){
	slot = 0;
	goto leave;
    }

    if( !_DPSendTo ){
	_DPSendTo = slot;
    }

    // Tell the server to clear any old entry it might have had...

    dpclient_message m;
    DWORD            cbWritten;
    int              tries;

    DPFillHeader( &m.header, DPCLIENT_TYPE_INIT );

    m.data[0] = 0;

    for( tries = 5; tries > 0; --tries ){
	if( WriteFile( slot, &m, sizeof( m ), &cbWritten, NULL ) ) break;
	Sleep( _DP_TIMEOUT );
    }

  leave:
    LeaveCriticalSection( &_DPCrit );
    return slot;
}

BOOL DPWrite( HANDLE slot, dpclient_message *msg )
{
    DWORD  cbWritten;
    BOOL   ok = FALSE;

    if( !slot ) slot = _DPSendTo;
    if( !slot ) return FALSE;

    EnterCriticalSection( &_DPCrit );

    if( msg ){
	for( int tries=5; tries > 0; --tries ){
	    if( WriteFile( slot, msg, sizeof( *msg ), &cbWritten, NULL ) ){
		ok = TRUE;
		goto leave;
            }
	    Sleep( _DP_TIMEOUT );
	}
    }

  leave:
    LeaveCriticalSection( &_DPCrit );
    return ok;
}

BOOL DPThreadName( HANDLE slot, const char *threadName )
{
    dpclient_message m;
    static int       threadNumber = 0;
    static char      buf[100];
    BOOL             ok = FALSE;

    if( !slot ) slot = _DPSendTo;
    if( !slot ) return FALSE;

    EnterCriticalSection( &_DPCrit );

    DPFillHeader( &m.header, DPCLIENT_TYPE_THREADNAME );

    if( !threadName ){
	threadName = buf;
	sprintf( buf, "Thread %d", ++threadNumber );
    }

    m.header.messageSize = (unsigned short) strlen( threadName );
    strcpy( (char *) m.data, threadName );

    ok = DPWrite( slot, &m );

    LeaveCriticalSection( &_DPCrit );
    return ok;
}

BOOL DPConnected( void )
{
    return( _DPSendTo != 0 );
}

BOOL DPSendString( HANDLE slot, const char *string )
{
    return DPSendTypedString( slot, string, 0 );
}

BOOL DPSendTypedString( HANDLE slot, const char *string, USHORT type )
{
    dpclient_message m;
    unsigned long    len;
    BOOL             ok = FALSE;
    DWORD            cbWritten;
    int              tries;

    if( !slot ) slot = _DPSendTo;
    if( !slot ) return FALSE;

    EnterCriticalSection( &_DPCrit );

    if( !string ) string  = "";

    DPFillHeader( &m.header, DPCLIENT_TYPE_STRING );

    len = strlen( string );

    while( len >= _MAX_DPMESSAGE_SIZE ){
	m.header.type        = DPCLIENT_TYPE_CONCAT | DPCLIENT_TYPE_STRING | type;
	m.header.messageSize = _MAX_DPMESSAGE_SIZE;
        m.header.messageID   = _DPNext++;

	memcpy( m.data, string, _MAX_DPMESSAGE_SIZE - 1 );
	m.data[ _MAX_DPMESSAGE_SIZE - 1 ] = 0;

        for( tries=5; tries > 0; --tries ){
	    ok = WriteFile( slot, &m, sizeof( m ), &cbWritten, NULL );
	    if( ok ) break;
	    Sleep( _DP_TIMEOUT );
	}

	if( !ok ) goto leave;

	string += _MAX_DPMESSAGE_SIZE - 1;
	len    -= _MAX_DPMESSAGE_SIZE - 1;
    }

    m.header.type        = DPCLIENT_TYPE_STRING | type;
    m.header.messageSize = (unsigned short) len + 1;
    m.header.messageID   = _DPNext++;

    memcpy( m.data, string, len + 1 );

    for( tries=5; tries > 0; --tries ){
	ok = WriteFile( slot, &m, sizeof( m ), &cbWritten, NULL );
	if( ok ) break;
	Sleep( _DP_TIMEOUT );
    }

  leave:
    LeaveCriticalSection( &_DPCrit );
    return ok;
}

BOOL DPClose( HANDLE handle )
{
    BOOL ok = FALSE;

    if( _DPNext == 0 ){
	DPInit();
    }

    EnterCriticalSection( &_DPCrit );

    if( !handle && _DPSendTo ){
	handle = _DPSendTo;
	_DPSendTo = 0;
    } else if( handle == _DPSendTo ){
	_DPSendTo = 0;
    }

    // Tell the server to clear any old entry it might have had...

    if( handle ){
	dpclient_message m;
	DWORD            cbWritten;
    
	DPFillHeader( &m.header, DPCLIENT_TYPE_FINI );

	m.data[0] = 0;

	for( int tries=5; tries > 0; --tries ){
	    if( WriteFile( handle, &m, sizeof( m ), &cbWritten, NULL ) ) break;
	    Sleep( _DP_TIMEOUT );
        }

	ok = CloseHandle( handle );
    }

    LeaveCriticalSection( &_DPCrit );
    return ok;
}

#endif

#pragma pack(pop);

#ifdef __cplusplus
};
#endif

#endif
